package io.tolgee.service.security

import io.tolgee.configuration.tolgee.TolgeeProperties
import io.tolgee.constants.Caches
import io.tolgee.constants.Message
import io.tolgee.dtos.cacheable.UserAccountDto
import io.tolgee.dtos.request.UserUpdatePasswordRequestDto
import io.tolgee.dtos.request.UserUpdateRequestDto
import io.tolgee.dtos.request.auth.SignUpDto
import io.tolgee.dtos.request.organization.OrganizationDto
import io.tolgee.dtos.request.validators.exceptions.ValidationException
import io.tolgee.events.user.OnUserCreated
import io.tolgee.events.user.OnUserUpdated
import io.tolgee.exceptions.AuthenticationException
import io.tolgee.exceptions.BadRequestException
import io.tolgee.exceptions.NotFoundException
import io.tolgee.exceptions.PermissionException
import io.tolgee.model.UserAccount
import io.tolgee.model.views.UserAccountInProjectView
import io.tolgee.model.views.UserAccountInProjectWithLanguagesView
import io.tolgee.model.views.UserAccountWithOrganizationRoleView
import io.tolgee.repository.UserAccountRepository
import io.tolgee.service.AvatarService
import io.tolgee.service.EmailVerificationService
import io.tolgee.service.organization.OrganizationService
import io.tolgee.util.executeInNewTransaction
import org.apache.commons.lang3.time.DateUtils
import org.hibernate.validator.internal.constraintvalidators.bv.EmailValidator
import org.springframework.beans.factory.annotation.Autowired
import org.springframework.cache.annotation.CacheEvict
import org.springframework.cache.annotation.Cacheable
import org.springframework.context.ApplicationEventPublisher
import org.springframework.context.annotation.Lazy
import org.springframework.data.domain.Page
import org.springframework.data.domain.Pageable
import org.springframework.security.crypto.password.PasswordEncoder
import org.springframework.stereotype.Service
import org.springframework.transaction.PlatformTransactionManager
import org.springframework.transaction.annotation.Transactional
import java.io.InputStream
import java.util.*
import javax.persistence.EntityManager

@Service
class UserAccountService(
  private val userAccountRepository: UserAccountRepository,
  private val applicationEventPublisher: ApplicationEventPublisher,
  private val tolgeeProperties: TolgeeProperties,
  private val avatarService: AvatarService,
  private val passwordEncoder: PasswordEncoder,
  @Lazy
  private val organizationService: OrganizationService,
  private val transactionManager: PlatformTransactionManager,
  private val entityManager: EntityManager
) {
  @Autowired
  lateinit var emailVerificationService: EmailVerificationService

  @Autowired
  @Lazy
  lateinit var permissionService: PermissionService

  private val emailValidator = EmailValidator()

  fun find(username: String): UserAccount? {
    return userAccountRepository.findNotDeleted(username)
  }

  operator fun get(username: String): UserAccount {
    return this.find(username) ?: throw NotFoundException(Message.USER_NOT_FOUND)
  }

  fun find(id: Long): UserAccount? {
    return userAccountRepository.findNotDeleted(id)
  }

  fun get(id: Long): UserAccount {
    return this.find(id) ?: throw NotFoundException(Message.USER_NOT_FOUND)
  }

  @Cacheable(cacheNames = [Caches.USER_ACCOUNTS], key = "#id")
  fun findDto(id: Long): UserAccountDto? {
    return userAccountRepository.findNotDeleted(id)?.let {
      UserAccountDto.fromEntity(it)
    }
  }

  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun createUser(userAccount: UserAccount): UserAccount {
    userAccountRepository.saveAndFlush(userAccount)
    applicationEventPublisher.publishEvent(OnUserCreated(this, userAccount))
    return userAccount
  }

  fun createUser(request: SignUpDto, role: UserAccount.Role = UserAccount.Role.USER): UserAccount {
    dtoToEntity(request).let {
      it.role = role
      this.createUser(it)
      return it
    }
  }

  @CacheEvict(Caches.USER_ACCOUNTS, key = "#id")
  @Transactional
  fun delete(id: Long) {
    val user = this.get(id)
    delete(user)
  }

  @CacheEvict(Caches.USER_ACCOUNTS, key = "#userAccount.id")
  @Transactional
  fun delete(userAccount: UserAccount) {
    userAccount.emailVerification?.let {
      entityManager.remove(it)
    }
    userAccount.apiKeys?.forEach {
      entityManager.remove(it)
    }
    userAccount.pats?.forEach {
      entityManager.remove(it)
    }
    userAccount.permissions.forEach {
      entityManager.remove(it)
    }
    userAccount.preferences?.let {
      entityManager.remove(it)
    }
    organizationService.getAllSingleOwnedByUser(userAccount).forEach {
      it.prefereredBy.removeIf { preferences ->
        preferences.userAccount.id == userAccount.id
      }
      organizationService.delete(it.id)
    }
    userAccount.organizationRoles.forEach {
      entityManager.remove(it)
    }
    userAccountRepository.softDeleteUser(userAccount)
  }

  fun dtoToEntity(request: SignUpDto): UserAccount {
    val encodedPassword = passwordEncoder.encode(request.password!!)
    return UserAccount(name = request.name, username = request.email, password = encodedPassword)
  }

  @get:Cacheable(cacheNames = [Caches.USER_ACCOUNTS], key = "'implicit'")
  val implicitUser: UserAccount
    get() {
      val username = "___implicit_user"
      return executeInNewTransaction(transactionManager) {
        userAccountRepository.findByUsername(username).orElseGet {
          val account = UserAccount(name = "No auth user", username = username, role = UserAccount.Role.ADMIN)
          this.createUser(account)
          this.organizationService.create(OrganizationDto(name = account.name), account)
          account
        }
      }
    }

  fun findByThirdParty(type: String, id: String): Optional<UserAccount> {
    return userAccountRepository.findThirdByThirdParty(id, type)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun setAccountType(userAccount: UserAccount, accountType: UserAccount.AccountType): UserAccount {
    userAccount.accountType = accountType
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun setResetPasswordCode(userAccount: UserAccount, code: String?): UserAccount {
    userAccount.resetPasswordCode = passwordEncoder.encode(code)
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun setUserPassword(userAccount: UserAccount, password: String?): UserAccount {
    userAccount.tokensValidNotBefore = DateUtils.truncate(Date(), Calendar.SECOND)
    userAccount.password = passwordEncoder.encode(password)
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  fun isResetCodeValid(userAccount: UserAccount, code: String?): Boolean {
    return passwordEncoder.matches(code, userAccount.resetPasswordCode)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun removeResetCode(userAccount: UserAccount): UserAccount {
    userAccount.resetPasswordCode = null
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun enableMfaTotp(userAccount: UserAccount, key: ByteArray): UserAccount {
    userAccount.totpKey = key
    userAccount.tokensValidNotBefore = DateUtils.truncate(Date(), Calendar.SECOND)
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun disableMfaTotp(userAccount: UserAccount): UserAccount {
    userAccount.totpKey = null
    // note: if support for more MFA methods is added, this should be only done if no other MFA method is enabled
    userAccount.mfaRecoveryCodes = emptyList()
    userAccount.tokensValidNotBefore = DateUtils.truncate(Date(), Calendar.SECOND)
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun consumeMfaRecoveryCode(userAccount: UserAccount, token: String): UserAccount {
    if (!userAccount.mfaRecoveryCodes.contains(token)) {
      throw AuthenticationException(Message.INVALID_OTP_CODE)
    }

    userAccount.mfaRecoveryCodes = userAccount.mfaRecoveryCodes.minus(token)
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun setMfaRecoveryCodes(userAccount: UserAccount, codes: List<String>): UserAccount {
    userAccount.mfaRecoveryCodes = codes
    return userAccountRepository.save(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#userAccount.id")
  fun removeAvatar(userAccount: UserAccount) {
    avatarService.removeAvatar(userAccount)
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#userAccount.id")
  fun setAvatar(userAccount: UserAccount, avatar: InputStream) {
    avatarService.setAvatar(userAccount, avatar)
  }

  fun getAllInOrganization(
    organizationId: Long,
    pageable: Pageable,
    search: String?
  ): Page<UserAccountWithOrganizationRoleView> {
    return userAccountRepository.getAllInOrganization(organizationId, pageable, search = search ?: "")
  }

  fun getAllInProject(
    projectId: Long,
    pageable: Pageable,
    search: String?,
    exceptUserId: Long? = null
  ): Page<UserAccountInProjectView> {
    return userAccountRepository.getAllInProject(projectId, pageable, search = search, exceptUserId)
  }

  fun getAllInProjectWithPermittedLanguages(
    projectId: Long,
    pageable: Pageable,
    search: String?,
    exceptUserId: Long? = null
  ): Page<UserAccountInProjectWithLanguagesView> {
    val users = getAllInProject(projectId, pageable, search, exceptUserId)
    val permittedLanguageMap = permissionService.getPermittedTranslateLanguagesForUserIds(
      users.content.map { it.id },
      projectId
    )
    return users.map {
      UserAccountInProjectWithLanguagesView(
        id = it.id,
        name = it.name,
        username = it.username,
        organizationRole = it.organizationRole,
        organizationBasePermissions = it.organizationBasePermissions,
        directPermissions = it.directPermissions,
        permittedLanguageIds = permittedLanguageMap[it.id]
      )
    }
  }

  @Transactional
  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun update(userAccount: UserAccount, dto: UserUpdateRequestDto): UserAccount {
    // Current password required to change email or password
    if (dto.email != userAccount.username) {
      if (userAccount.accountType == UserAccount.AccountType.LDAP) {
        throw BadRequestException(Message.OPERATION_UNAVAILABLE_FOR_ACCOUNT_TYPE)
      }

      if (dto.currentPassword?.isNotEmpty() != true) throw BadRequestException(Message.CURRENT_PASSWORD_REQUIRED)
      val matches = passwordEncoder.matches(dto.currentPassword, userAccount.password)
      if (!matches) throw BadRequestException(Message.WRONG_CURRENT_PASSWORD)
    }

    val old = UserAccountDto.fromEntity(userAccount)
    updateUserEmail(userAccount, dto)
    userAccount.name = dto.name

    publishUserInfoUpdatedEvent(old, userAccount)
    return userAccountRepository.save(userAccount)
  }

  fun updatePassword(userAccount: UserAccount, dto: UserUpdatePasswordRequestDto): UserAccount {
    if (userAccount.accountType == UserAccount.AccountType.LDAP) {
      throw BadRequestException(Message.OPERATION_UNAVAILABLE_FOR_ACCOUNT_TYPE)
    }

    val matches = passwordEncoder.matches(dto.currentPassword, userAccount.password)
    if (!matches) throw PermissionException()

    userAccount.tokensValidNotBefore = DateUtils.truncate(Date(), Calendar.SECOND)
    userAccount.password = passwordEncoder.encode(dto.password)
    return userAccountRepository.save(userAccount)
  }

  private fun updateUserEmail(
    userAccount: UserAccount,
    dto: UserUpdateRequestDto
  ) {
    if (userAccount.username != dto.email) {
      if (!emailValidator.isValid(dto.email, null)) {
        // todo: Allow to specify STANDARD_VALIDATION typed errors to show errors on specific fields
        throw ValidationException(Message.VALIDATION_EMAIL_IS_NOT_VALID)
      }

      this.find(dto.email)?.let { throw ValidationException(Message.USERNAME_ALREADY_EXISTS) }
      if (tolgeeProperties.authentication.needsEmailVerification) {
        emailVerificationService.createForUser(userAccount, dto.callbackUrl, dto.email)
      } else {
        userAccount.username = dto.email
      }
    }
  }

  private fun publishUserInfoUpdatedEvent(
    old: UserAccountDto,
    userAccount: UserAccount
  ) {
    val event = OnUserUpdated(this, old, UserAccountDto.fromEntity(userAccount))
    applicationEventPublisher.publishEvent(event)
  }

  fun saveAll(userAccounts: Collection<UserAccount>): MutableList<UserAccount> =
    userAccountRepository.saveAll(userAccounts)

  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun save(user: UserAccount): UserAccount {
    return userAccountRepository.save(user)
  }

  @CacheEvict(cacheNames = [Caches.USER_ACCOUNTS], key = "#result.id")
  fun saveAndFlush(user: UserAccount): UserAccount {
    return userAccountRepository.saveAndFlush(user)
  }

  fun getAllByIdsIncludingDeleted(ids: Set<Long>): MutableList<UserAccount> {
    return userAccountRepository.getAllByIdsIncludingDeleted(ids)
  }

  fun findAllPaged(pageable: Pageable, search: String?): Page<UserAccount> {
    return userAccountRepository.findAllPaged(search, pageable)
  }

  val isAnyUserAccount: Boolean
    get() = userAccountRepository.count() > 0
}
